/********************************************************************\

  Name:         frontend.c
  Created by:   Stefan Ritt

  Contents:     Experiment specific readout code (user part) of
                Midas frontend. This example simulates a "trigger
                event" and a "scaler event" which are filled with
                random data. The trigger event is filled with
                two banks (ADC0 and TDC0), the scaler event with 
                one bank (SCLR).


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  25-JAN-95    SR    created


\********************************************************************/

#include <stdio.h>
#include "midas.h"
#include "event.h"

/* make frontend functions callable from the C framework */
#ifdef __cplusplus
extern "C" {
#endif

/*-- Globals -------------------------------------------------------*/

/* The frontend name (client name) as seen by other MIDAS clients   */
char *frontend_name = "Trigger Frontend";

/* frontend_loop is called periodically if this variable is TRUE    */
BOOL frontend_call_loop = FALSE;

/* a frontend status page is displayed with this frequency in ms */
INT display_period = 3000;

/*-- Function declarations -----------------------------------------*/

INT frontend_init();
INT frontend_exit();
INT begin_of_run(INT run_number, char *error);
INT end_of_run(INT run_number, char *error);
INT pause_run(INT run_number, char *error);
INT resume_run(INT run_number, char *error);
INT frontend_loop();

INT poll_trigger_event(INT count, PTYPE test);
INT read_trigger_event(char *pevent);
INT read_scaler_event(char *pevent);

/*-- Equipment list ------------------------------------------------*/

EQUIPMENT equipment[] = {

  { "Trigger",            /* equipment name */
    1, 0,                 /* event ID, trigger mask */
    "SYSTEM",             /* event buffer */
    EQ_POLLED,            /* equipment type */
    "MIDAS",              /* format */
    TRUE,                 /* enabled */
    RO_RUNNING |          /* read only when running */
    RO_ODB,               /* and update ODB */ 
    100,                  /* poll for 100ms */
    0,                    /* stop run after this event limit */
    0,                    /* don't log history */
    "", "", 
    read_trigger_event,   /* readout routine */
    poll_trigger_event,   /* polling routine */
    NULL,                 /* init string */
  },

  { "Scaler",             /* equipment name */
    2, 0,                 /* event ID, trigger mask */
    "SYSTEM",             /* event buffer */
    EQ_PERIODIC,          /* equipment type */
    "MIDAS",              /* format */
    TRUE,                 /* enabled */
    RO_RUNNING |
    RO_TRANSITIONS |      /* read when running and on transitions */
    RO_ODB,               /* and update ODB */ 
    10000,                /* read every 10 sec */
    0,                    /* stop run after this event limit */
    0,                    /* log history */
    "", "", 
    read_scaler_event,    /* readout routine */
    NULL,                 /* polling routine */
    NULL,                 /* init string */
  },

  { "" }
};

#ifdef __cplusplus
}
#endif

/********************************************************************\
              Callback routines for system transitions

  These routines are called whenever a system transition like start/
  stop of a run occurs. The routines are called on the following
  occations:

  frontend_init:  When the frontend program is started. This routine
                  should initialize the hardware.
  
  frontend_exit:  When the frontend program is shut down. Can be used
                  to releas any locked resources like memory, commu-
                  nications ports etc.

  begin_of_run:   When a new run is started. Clear scalers, open
                  rungates, etc.

  end_of_run:     Called on a request to stop a run. Can send 
                  end-of-run event and close run gates.

  pause_run:      When a run is paused. Should disable trigger events.

  resume_run:     When a run is resumed. Should enable trigger events.

\********************************************************************/

/*-- Frontend Init -------------------------------------------------*/

INT frontend_init()
{
  /* put here hardware initialization */

  return CM_SUCCESS;
}

/*-- Frontend Exit -------------------------------------------------*/

INT frontend_exit()
{
  return CM_SUCCESS;
}

/*-- Begin of Run --------------------------------------------------*/

INT begin_of_run(INT run_number, char *error)
{
  /* put here clear scalers etc. */

  return CM_SUCCESS;
}

/*-- End of Run ----------------------------------------------------*/

INT end_of_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Pause Run -----------------------------------------------------*/

INT pause_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Resuem Run ----------------------------------------------------*/

INT resume_run(INT run_number, char *error)
{
  return CM_SUCCESS;
}

/*-- Frontend Loop -------------------------------------------------*/

INT frontend_loop()
{
  /* if frontend_call_loop is true, this routine gets called when
     the frontend is idle or between every event */
  return CM_SUCCESS;
}

/*------------------------------------------------------------------*/

/********************************************************************\
  
  Readout routines for different events

\********************************************************************/

/*-- Trigger event -------------------------------------------------*/

INT poll_trigger_event(INT count, PTYPE test)
/* Polling routine for trigger event. Return TRUE if event
   is available. If test equals TRUE, don't return. The test
   flag is used to time the polling */
{
INT i;

  for (i=0 ; i<count ; i++)
    {
    if (i == 0)      /* put here if("event available") hardware test */
      if (!test)
        return TRUE;
    }

  return FALSE;
}


INT read_trigger_event(char *pevent)
{
DWORD *pdata, i;

  /* init bank structure */
  bk_init(pevent);

  /* create ADC bank */
  bk_create(pevent, "ADC0", TID_DWORD, &pdata);

  /* fill ADC bank with random data */
  for (i=0 ; i<8 ; i++)
    *pdata++ = rand() & 0xFFF;

  bk_close(pevent, pdata);

  /* create TDC bank */
  bk_create(pevent, "TDC0", TID_DWORD, &pdata);

  /* fill TDC bank with random data */
  for (i=0 ; i<8 ; i++)
    *pdata++ = rand() & 0xFFF;

  bk_close(pevent, pdata);

  /* simulate idle time to reduce data rate */
  /*  ss_sleep(10); */

  return bk_size(pevent);
}

/*-- Scaler event --------------------------------------------------*/

INT read_scaler_event(char *pevent)
{
DWORD *pdata, i;

  /* init bank structure */
  bk_init(pevent);

  /* create SCLR bank */
  bk_create(pevent, "SCLR", TID_DWORD, &pdata);

  /* fill scaler bank with random data */
  for (i=0 ; i<8 ; i++)
    *pdata++ = rand();

  bk_close(pevent, pdata);

  return bk_size(pevent);
}
