/********************************************************************\

  Name:         mfe.c
  Created by:   Stefan Ritt

  Contents:     The system part of the MIDAS frontend. Has to be
                linked with mfeuser.c to form a complete frontend


  Revision history
  ------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  24-SEP-96    SR    created


\********************************************************************/

#include <stdio.h>
#include "midas.h"
#include "msystem.h"
#include "ybos.h"

/*------------------------------------------------------------------*/

/* items defined in frontend.c */

extern char *frontend_name;
extern BOOL frontend_call_loop;
extern INT  display_period;
extern INT  frontend_init(void);
extern INT  frontend_exit(void);
extern INT  frontend_loop(void);
extern INT  begin_of_run(INT run_number, char *error);
extern INT  end_of_run(INT run_number, char *error);
extern INT  pause_run(INT run_number, char *error);
extern INT  resume_run(INT run_number, char *error);

/*------------------------------------------------------------------*/

/* globals */

INT   run_state;      /* STATE_RUNNING, STATE_STOPPED, STATE_PAUSED */
INT   run_number;
char  *event_buffer;
char  host_name[NAME_LENGTH];
char  exp_name[NAME_LENGTH];
HNDLE hDB;

struct {
  DWORD ybos_type;
  DWORD odb_type;
  INT tsize;
} id_map[] = {
  {A1_BKTYPE, TID_CHAR, 1},
  {I1_BKTYPE, TID_BYTE, 1},
  {I2_BKTYPE, TID_WORD, 2},
  {I4_BKTYPE, TID_DWORD, 4},
  {F4_BKTYPE, TID_FLOAT, 4},
  {0,0}
};

extern EQUIPMENT equipment[];

void send_all_periodic_events(INT transition);

/*---- ODB records -------------------------------------------------*/

#define EQUIPMENT_COMMON_STR "\
Event ID = WORD : 0\n\
Trigger mask = WORD : 0\n\
Buffer = STRING : [32] SYSTEM\n\
Type = INT : 0\n\
Format = STRING : [8] FIXED\n\
Enabled = BOOL : 0\n\
Read on = INT : 0\n\
Period = INT : 0\n\
Event limit = DWORD : 0\n\
Log history = INT : 0\n\
Client name = STRING : [32] \n\
Host = STRING : [32] \n\
"

#define EQUIPMENT_STATISTICS_STR "\
Events sent = DWORD : 0\n\
Events per sec. = DWORD : 0\n\
kBytes per sec. = DOUBLE : 0\n\
"

/*-- transition callbacks ------------------------------------------*/

/*-- start ---------------------------------------------------------*/

INT tr_start(INT rn, char *error)
{
INT i, status;

  /* reset serial numbers */
  for (i=0 ; equipment[i].eq_name[0] ; i++)
    {
    equipment[i].serial_number = 1;
    equipment[i].odb_in = equipment[i].odb_out = 0;
    }

  status = begin_of_run(rn, error);

  if (status == CM_SUCCESS)
    {
    run_state = STATE_RUNNING;
    run_number = rn;
    }

  send_all_periodic_events(TR_START);

  if (display_period)
    {
    ss_printf(14, 2, "Running ");
    ss_printf(36, 2, "%d", rn);
    }

  return status;
}

/*-- prestop -------------------------------------------------------*/

INT tr_prestop(INT rn, char *error)
{
INT status;

  status = end_of_run(rn, error);

  /* don't send events if already stopped */
  if (run_state != STATE_STOPPED)
    send_all_periodic_events(TR_STOP);

  if (status == CM_SUCCESS)
    {
    run_state = STATE_STOPPED;
    run_number = rn;
    }

  if (display_period)
    ss_printf(14, 2, "Stopped ");

  return status;
}

/*-- pause ---------------------------------------------------------*/

INT tr_pause(INT rn, char *error)
{
INT status;

  status = pause_run(rn, error);

  if (status == CM_SUCCESS)
    {
    run_state = STATE_PAUSED;
    run_number = rn;
    }

  send_all_periodic_events(TR_PAUSE);

  if (display_period)
    ss_printf(14, 2, "Paused  ");

  return status;
}

/*-- resume --------------------------------------------------------*/

INT tr_resume(INT rn, char *error)
{
INT status;

  status = resume_run(rn, error);

  if (status == CM_SUCCESS)
    {
    run_state = STATE_RUNNING;
    run_number = rn;
    }

  send_all_periodic_events(TR_RESUME);

  if (display_period)
    ss_printf(14, 2, "Running ");

  return status;
}

/*------------------------------------------------------------------*/

void register_equipment(void)
{
INT    index, count, size, status;
char   str[256];
EQUIPMENT_INFO *eq_info;
EQUIPMENT_STATS *eq_stats;
DWORD  start_time, delta_time;
HNDLE  hKey;

  /* get current ODB run state */
  size = sizeof(run_state);
  run_state = STATE_STOPPED;
  db_get_value(hDB, 0, "/Runinfo/State", &run_state, &size, TID_INT);
  size = sizeof(run_number);
  run_number = 1;
  db_get_value(hDB, 0, "/Runinfo/Run number", &run_number, &size, TID_INT);

  /* scan EQUIPMENT table from FRONTEND.C */
  for (index=0 ; equipment[index].eq_name[0] ; index++)
    {
    eq_info = &equipment[index].eq_info;
    eq_stats = &equipment[index].eq_stats;

    if (eq_info->event_id == 0)
      {
      printf("Event ID 0 for %s not allowed\n", equipment[index].eq_name);
      ss_sleep(5000);
      }

    sprintf(str, "/Equipment/%s/Common", equipment[index].eq_name);
    
    /* get last event limit from ODB */
    db_find_key(hDB, 0, str, &hKey);
    if (hKey)
      db_get_value(hDB, hKey, "Event limit", &eq_info->event_limit, &size, TID_DWORD);
    
    /* Create common subtree */
    status = db_create_record(hDB, 0, str, EQUIPMENT_COMMON_STR);
    if (status != DB_SUCCESS)
      {
      printf("Cannot init equipment record, probably other FE is using it\n");
      ss_sleep(3000);
      }
    db_find_key(hDB, 0, str, &hKey);

    if (equal_ustring(eq_info->format, "YBOS"))
      equipment[index].format = FORMAT_YBOS;
    else if (equal_ustring(eq_info->format, "FIXED"))
      equipment[index].format = FORMAT_FIXED;
    else /* default format is MIDAS */
      equipment[index].format = FORMAT_MIDAS;

    gethostname(eq_info->host, sizeof(eq_info->host));
    strcpy(eq_info->client_name, frontend_name);

    /* set record from equipment[] table in frontend.c */
    db_set_record(hDB, hKey, eq_info, sizeof(EQUIPMENT_INFO), 0);

    /* open hot link to equipment info */
    db_open_record(hDB, hKey, eq_info, sizeof(EQUIPMENT_INFO), MODE_READ, NULL);

    /*---- Create variables record ---------------------------------*/
    sprintf(str, "/Equipment/%s/Variables", equipment[index].eq_name);
    if (equipment[index].init_string)
      db_create_record(hDB, 0, str, equipment[index].init_string);
    else
      db_create_key(hDB, 0, str, TID_KEY);
    db_find_key(hDB, 0, str, &hKey);
    equipment[index].hkey_variables = hKey;

    /*---- Create and initialize statistics tree -------------------*/
    sprintf(str, "/Equipment/%s/Statistics", equipment[index].eq_name);
    db_create_record(hDB, 0, str, EQUIPMENT_STATISTICS_STR);
    db_find_key(hDB, 0, str, &hKey);

    eq_stats->events_sent = 0;
    eq_stats->events_per_sec = 0;
    eq_stats->kbytes_per_sec = 0;

    /* open hot link to statistics tree */
    status = db_open_record(hDB, hKey, eq_stats, sizeof(EQUIPMENT_STATS), MODE_WRITE, NULL);
    if (status != DB_SUCCESS)
      {
      printf("Cannot open statistics record, probably other FE is using it\n");
      ss_sleep(3000);
      }

    /*---- open event buffer ---------------------------------------*/
    if (eq_info->buffer[0])
      {
      bm_open_buffer(eq_info->buffer, EVENT_BUFFER_SIZE, &equipment[index].buffer_handle);

      /* set the default buffer cache size */
      bm_set_cache_size(equipment[index].buffer_handle, 0, 100000);
      }
    else
      equipment[index].buffer_handle = 0;

    /*---- evaluate polling count ----------------------------------*/
    if (eq_info->eq_type == EQ_POLLED)
      {
      if (display_period)
        printf("\nCalibrating");

      count = 1;
      do
        {
        if (display_period)
          printf(".");

        start_time = ss_millitime();

        equipment[index].func(count, TRUE);

        delta_time = ss_millitime() - start_time;

        if (delta_time > 0)
          count = (INT) ((double) count * 100 / delta_time);
        else
          count*=100;
        } while (delta_time > 120 || delta_time < 80);

      equipment[index].poll_count = (INT) ((double) eq_info->period / 100 * count);

      if (display_period)
        printf("OK\n");
      }

    /*---- initialize slow control equipment -----------------------*/
    if (eq_info->eq_type == EQ_SLOW)
      {
      /* loop over equipment list and call class driver's init method */
      if (eq_info->enabled)
        equipment[index].status = equipment[index].func(CMD_INIT, (PTYPE) &equipment[index]);
      else
        {
        equipment[index].status = FE_ERR_DISABLED;
        cm_msg(MINFO, "register_equipment", "Equipment %s disabled in file \"frontend.c\"",
                       equipment[index].eq_name);
        }

      /* let user read error messages */
      if (equipment[index].status != FE_SUCCESS)
        ss_sleep(3000);
      }
    }
}

/*------------------------------------------------------------------*/

void update_odb(EVENT_HEADER *pevent, HNDLE hKey, INT format)
{
INT               size, i, n, ni4, tsize, status;
void              *pdata;
char              name[5];
BANK_HEADER       *pbh;
BANK              *pbk;
void              *pydata;
DWORD             odb_type;
DWORD             *pyevt;
YBOS_BANK_HEADER  *pybkh;

  if (format == FORMAT_FIXED)
    {
    if (db_set_record(hDB, hKey, (char *) (pevent+1),
                      pevent->data_size, 0) != DB_SUCCESS)
      cm_msg(MERROR, "update_odb", "event #%d size mismatch", pevent->event_id);
    }
  else if (format == FORMAT_MIDAS)
    {
    pbh = (BANK_HEADER *) (pevent+1);
    pbk = NULL;
    do
      {
      /* scan all banks */
      size = bk_iterate(pbh, &pbk, &pdata);
      if (pbk == NULL)
        break;
      
      /* write bank to ODB */
      *((DWORD *) name) = *((DWORD *) (pbk)->name);
      name[4] = 0;

      n = rpc_tid_size(pbk->type & 0xFF) ? size / rpc_tid_size(pbk->type & 0xFF) : size;

      if (n>0)
        db_set_value(hDB, hKey, name, pdata, size, n, pbk->type & 0xFF);

      } while (1);
    }
  else if (format == FORMAT_YBOS)
    {
    /* skip the lrl (4 bytes per event) */
    pyevt = (DWORD *) (pevent+1);
    pybkh = NULL;
    do
	    {
	    /* scan all banks */
	    ni4 = ybkiterate(pyevt, &pybkh, &pydata);
	    if (pybkh == NULL || ni4 == 0)
	      break;

	    /* find the corresponding odb type */
	    for (i=0;id_map[0].ybos_type > 0; i++)
	      {
        if (pybkh->type == id_map[i].ybos_type)
          {
		      odb_type = id_map[i].odb_type;
		      tsize = id_map[i].tsize;
 		      break;
          }
	      }

	    /* extract bank name (key name) */
	    *((DWORD *)name) = pybkh->name;
	    name[4] = 0;

      /* correct YBS number of entries */
      if (pybkh->type == I2_BKTYPE)
        ni4 *= 2;
      if (pybkh->type == I1_BKTYPE || pybkh->type == A1_BKTYPE)
        ni4 *= 4;

	    /* write bank to ODB, ni4 always in I*4 */
	    size = ni4 * tsize;
	    if ((status = db_set_value(hDB, hKey, name, pydata, size, ni4, odb_type & 0xFF)) != DB_SUCCESS)
	      {
	      printf("status:%i odb_type:%i name:%s ni4:%i size:%i tsize:%i\n",
		            status, odb_type, name, ni4, size, tsize);
	      for (i=0;i<6;i++)
		      printf("data: %f\n",*((float *)(pydata))++);
	      }
      } while (1);
    }
}

/*------------------------------------------------------------------*/

void send_all_periodic_events(INT transition)
{
EQUIPMENT_INFO *eq_info;
EVENT_HEADER   *pevent;
DWORD          actual_time;
INT            i;

  pevent = (EVENT_HEADER *) event_buffer;

  for (i=0 ; equipment[i].eq_name[0] ; i++)
    {
    eq_info = &equipment[i].eq_info;

    if ((eq_info->eq_type != EQ_PERIODIC && eq_info->eq_type != EQ_SLOW) || 
        !eq_info->enabled)
      continue;

    if (transition == TR_START  && (eq_info->read_on & RO_BOR)    == 0)
      continue;
    if (transition == TR_STOP   && (eq_info->read_on & RO_EOR)    == 0)
      continue;
    if (transition == TR_PAUSE  && (eq_info->read_on & RO_PAUSE)  == 0)
      continue;
    if (transition == TR_RESUME && (eq_info->read_on & RO_RESUME) == 0)
      continue;

    actual_time = ss_millitime();

    pevent->event_id      = eq_info->event_id;
    pevent->trigger_mask  = eq_info->trigger_mask;
    pevent->data_size     = 0;
    pevent->time_stamp    = ss_millitime();
    pevent->serial_number = equipment[i].serial_number++;

    equipment[i].last_called = actual_time;

    /* call user readout routine */
    pevent->data_size = equipment[i].readout((char *) (pevent+1));

    /* send event */
    if (pevent->data_size)
      {
      equipment[i].bytes_sent += pevent->data_size;
      equipment[i].events_sent++;

      /* send event to buffer */
      if (equipment[i].buffer_handle)
        {
        bm_send_event(equipment[i].buffer_handle, pevent, 
                      pevent->data_size + sizeof(EVENT_HEADER), SYNC);

        /* flush buffer cache */
        bm_flush_cache(equipment[i].buffer_handle, SYNC);
        }

      /* send event to ODB */
      if (eq_info->read_on & RO_ODB ||
          eq_info->history > 0)
        {
        update_odb(pevent, equipment[i].hkey_variables, equipment[i].format);
        equipment[i].odb_out++;
        }
      }
    else
      equipment[i].serial_number--;
    }
}

/*------------------------------------------------------------------*/

int message_print(const char *msg)
{
char str[160];

  memset(str, ' ', 159);
  str[159] = 0;

  if (msg[0] == '[')
    msg = strchr(msg, ']')+2;

  memcpy(str, msg, strlen(msg));
  ss_printf(0, 20, str);

  return 0;
}

void display(BOOL bInit)
{
INT    i, status;
time_t full_time;
char   str[30];

  if (bInit)
    {
    ss_clear_screen();

    if (host_name[0])
      strcpy(str, host_name);
    else
      strcpy(str, "<local>");

    ss_printf(0, 0, "%s connected to %s. Press \"!\" to exit", frontend_name, str);
    ss_printf(0, 1, "================================================================================");
    ss_printf(0, 2, "Run status:   %s", run_state == STATE_STOPPED ? "Stopped" :
                                        run_state == STATE_RUNNING ? "Running" :
                                        "Paused");
    ss_printf(25,2, "Run number %ld   ", run_number);
    ss_printf(0, 3, "================================================================================");
    ss_printf(0, 4, "Equipment     Status     Events     Events/sec Rate[kB/s] ODB->FE    FE->ODB");
    ss_printf(0, 5, "--------------------------------------------------------------------------------");
    for (i=0 ; equipment[i].eq_name[0] ; i++)
      ss_printf(0, i+6, "%s", equipment[i].eq_name);
    }

  /* display time */
  time(&full_time);
  strcpy(str, ctime(&full_time)+11);
  str[8] = 0;
  ss_printf(72, 0, "%s", str);

  for (i=0 ; equipment[i].eq_name[0] ; i++)
    {
    status = equipment[i].status;

    if ((status == 0 || status == FE_SUCCESS) &&
        equipment[i].eq_info.enabled)
      ss_printf(14, i+6, "OK       ");
    else if (!equipment[i].eq_info.enabled)
      ss_printf(14, i+6, "Disabled ");
    else if (status == FE_ERR_ODB)
      ss_printf(14, i+6, "ODB Error");
    else if (status == FE_ERR_HW)
      ss_printf(14, i+6, "HW Error ");
    else if (status == FE_ERR_DISABLED)
      ss_printf(14, i+6, "Disabled ");
    else
      ss_printf(14, i+6, "Unknown  ");

    ss_printf(25, i+6, "%ld       ", equipment[i].eq_stats.events_sent);
    ss_printf(36, i+6, "%ld       ", equipment[i].eq_stats.events_per_sec);
    ss_printf(47, i+6, "%1.1lf      ", equipment[i].eq_stats.kbytes_per_sec);
    ss_printf(58, i+6, "%ld       ", equipment[i].odb_in);
    ss_printf(69, i+6, "%ld       ", equipment[i].odb_out);
    }
}

/*------------------------------------------------------------------*/

INT scheduler(void)
{
EQUIPMENT_INFO *eq_info;
EQUIPMENT      *eq;
EVENT_HEADER   *pevent;
DWORD          actual_time, last_time_network=0, last_time_display=0,
               readout_start;
INT            i, j, index, status, old_state, ch;
char           str[80];

  pevent = (EVENT_HEADER *) event_buffer;

  do
    {
    actual_time = ss_millitime();

    /*---- loop over equipment table -------------------------------*/
    for (index=0 ; ; index++)
      {
      eq = &equipment[index];
      eq_info = &eq->eq_info;

      /* check if end of equipment list */
      if (!eq->eq_name[0])
        break;

      if (!eq_info->enabled)
        continue;

      /*---- call idle routine for slow control equipment ----*/
      if (eq_info->eq_type == EQ_SLOW &&
          equipment[index].status == FE_SUCCESS)
        equipment[index].func(CMD_IDLE, (PTYPE) &equipment[index]);

      if (run_state == STATE_STOPPED && (eq_info->read_on & RO_STOPPED) == 0)
        continue;
      if (run_state == STATE_PAUSED  && (eq_info->read_on & RO_PAUSED)  == 0)
        continue;
      if (run_state == STATE_RUNNING && (eq_info->read_on & RO_RUNNING) == 0)
        continue;

      /*---- check periodic events ----*/
      if (eq_info->eq_type == EQ_PERIODIC || eq_info->eq_type == EQ_SLOW)
        {
        if (eq_info->period == 0)
          continue;

        /* if period over, call readout routine */
        if (actual_time - eq->last_called >= (DWORD) eq_info->period)
          {
          eq->last_called = actual_time;

          /* compose MIDAS event header */
          pevent->event_id      = eq_info->event_id;
          pevent->trigger_mask  = eq_info->trigger_mask;
          pevent->data_size     = 0;
          pevent->time_stamp    = actual_time;
          pevent->serial_number = eq->serial_number++;

          /* call user readout routine */
          pevent->data_size = eq->readout((char *) (pevent+1));

          /* send event */
          if (pevent->data_size)
            {
            eq->bytes_sent += pevent->data_size;
            eq->events_sent++;

            if (eq->buffer_handle)
              {
              bm_send_event(eq->buffer_handle, pevent,
                            pevent->data_size + sizeof(EVENT_HEADER), SYNC);

              /* flush buffer cache */
              bm_flush_cache(eq->buffer_handle, SYNC);
              }

            /* send event to ODB for periodic events */
            if (eq_info->eq_type == EQ_PERIODIC &&
                (eq_info->read_on & RO_ODB ||
                 eq_info->history > 0))
              {
              update_odb(pevent, eq->hkey_variables, eq->format);
              eq->odb_out++;
              }
            }
          else
            eq->serial_number--;
          }
        }

      /*---- check polled events ----*/
      if (eq_info->eq_type == EQ_POLLED)
        {
        /* compose MIDAS event header */
        pevent->event_id      = eq_info->event_id;
        pevent->trigger_mask  = eq_info->trigger_mask;
        pevent->data_size     = 0;
        readout_start = actual_time;

        while (eq->func(eq->poll_count, FALSE))
          {
          pevent->time_stamp = actual_time;
          pevent->serial_number = eq->serial_number++;

          /* call user readout routine */
          pevent->data_size = eq->readout((char *) (pevent+1));

          /* send event */
          if (pevent->data_size)
            {
            if (eq->buffer_handle)
              do
                {
                status = rpc_send_event(0, eq->buffer_handle, pevent,
                                        pevent->data_size + sizeof(EVENT_HEADER), SYNC);

                if (status != BM_SUCCESS && status != BM_ASYNC_RETURN)
                  {
                  cm_msg(MERROR, "scheduler", "error sending event");
                  return status;
                  }

                if (status == BM_ASYNC_RETURN)
                  {
                  old_state = run_state;

                  cm_yield(10);

                  /* check if run state has changed in between */
                  if (run_state != old_state)
                    break;
                  }

                } while (status == BM_ASYNC_RETURN);

            eq->bytes_sent += pevent->data_size;
            eq->events_sent++;
            }
          else
            eq->serial_number--;

          actual_time = ss_millitime();

          /* repeat no more than 100 ms */
          if (actual_time - readout_start > 100)
            break;

          /* quit if event limit is reached */
          if (eq_info->event_limit &&
              eq->serial_number > eq_info->event_limit)
            break;
          }

        /* send event to ODB */
        if (eq_info->read_on & RO_ODB ||
            eq_info->history)
          {
          if (actual_time - eq->last_called > 10000 && pevent->data_size)
            {
            eq->last_called = actual_time;
            update_odb(pevent, eq->hkey_variables, eq->format);
            eq->odb_out++;
            }
          }
        }

      /*---- check if event limit is reached ----*/
      if (eq_info->event_limit &&
          eq->serial_number > eq_info->event_limit &&
          run_state == STATE_RUNNING)
        {
        /* stop run */
        if (cm_transition(TR_STOP, 0, str, sizeof(str), SYNC) != CM_SUCCESS)
          {
          cm_msg(MERROR, "scheduler", "cannot stop run: %s", str);
          }
        }
      }

    /*---- call frontend_loop periodically -------------------------*/
    if (frontend_call_loop)
      frontend_loop();

    /*---- calculate rates and update status page periodically -----*/
    if ((display_period && actual_time - last_time_display > (DWORD) display_period) ||
        (!display_period && actual_time - last_time_display > 5000))
      {
      /* calculate rates */
      if (actual_time != last_time_display)
        for (i=0 ; equipment[i].eq_name[0] ; i++)
          {
          eq = &equipment[i];
          j = eq->serial_number;
          if (j > 0)
            j--;
          eq->eq_stats.events_sent = j;
          eq->eq_stats.events_per_sec =
            (DWORD) (eq->events_sent / ((actual_time - last_time_display)/1000.0));
          eq->eq_stats.kbytes_per_sec =
            eq->bytes_sent/1024.0 / ((actual_time - last_time_display)/1000.0);

          eq->bytes_sent = 0;
          eq->events_sent = 0;
          }

      /* propagate changes in equipment to ODB */
      db_send_changed_records();

      if (display_period)
        {
        display(FALSE);

        /* check keyboard */
        ch = 0;
        if (ss_kbhit())
          {
#if defined(OS_MSDOS) || defined(OS_WINNT)
          ch = getch();
#else
          ch = getchar();
#endif
          }

        /*
        ch = ss_getchar(0);
        if (ch == -1)
          {
          if (ss_kbhit())
            ch = getchar();
          }
        */

        if (ch == '!')
          {
          status = RPC_SHUTDOWN;
          break;
          }

        if (ch > 0)
          display(TRUE);
        }

      last_time_display = actual_time;
      }

    /*---- check network messages ----------------------------------*/
    if (run_state == STATE_RUNNING)
      {
      /* only call yield once every 100ms when running */
      if (actual_time - last_time_network > 100)
        {
        status = cm_yield(0);
        last_time_network = actual_time;
        }
      else
        status = RPC_SUCCESS;
      }
    else
      /* when run is stopped, call yield with 100ms timeout */
      status = cm_yield(100);

    } while (status != RPC_SHUTDOWN && status != SS_ABORT);

  return status;
}

/*------------------------------------------------------------------*/

#ifdef OS_VXWORKS
mfe(char *ahost_name, char *aexp_name, BOOL adebug)
#else
main(int argc, char *argv[])
#endif
{
INT    status, i;
BOOL   debug;

  host_name[0] = 0;
  exp_name[0] = 0;
  debug = FALSE;

#ifdef OS_VXWORKS
  if (ahost_name)
    strcpy(host_name, ahost_name);
  if (aexp_name)
    strcpy(exp_name, aexp_name);
  debug = adebug;
#else
  /* parse command line parameters */
  for (i=1 ; i<argc ; i++)
    {
    if (argv[i][0] == '-' && argv[i][1] == 'd')
      debug = TRUE;
    else if (argv[i][0] == '-')
      {
      if (i+1 >= argc || argv[i+1][0] == '-')
        goto usage;
      if (argv[i][1] == 'e')
        strcpy(exp_name, argv[++i]);
      else if (argv[i][1] == 'h')
        strcpy(host_name, argv[++i]);
      else
        {
usage:
        printf("usage: frontend [-h Hostname] [-e Experiment] [-d]\n\n");
        return 0;
        }
      }
    }
#endif

  if (equal_ustring(host_name, "local"))
    host_name[0] = 0;

  if (exp_name[0] == 0)
    {
    status = cm_select_experiment(host_name, exp_name);
    if (status != CM_SUCCESS)
      return 1;
    }

  /* allocate event buffer */
  event_buffer = malloc(MAX_EVENT_SIZE + sizeof(EVENT_HEADER));
  if (event_buffer == NULL)
    {
    printf("Not enough memeory to allocate event buffer\n");
    return 1;
    }

  /* now connect to server */
  if (display_period)
    {
    if (host_name[0])
      printf("Connect to experiment %s on host %s...", exp_name, host_name);
    else
      printf("Connect to experiment %s...", exp_name);
    }

  status = cm_connect_experiment(host_name, exp_name, frontend_name, NULL);
  
  if (status == CM_WRONG_PASSWORD)
    printf("Wrong password\n");
  else if (status == CM_UNDEF_EXP)
    printf("Experiment %s not defined\n", exp_name);
  else if (status == RPC_NET_ERROR)
    printf("Cannot connect to remote host %s\n", host_name);
  else if (status == CM_UNDEF_ENVIRON)
    printf("MIDAS environment variable not defined\n");
  else if (status != CM_SUCCESS)
    printf("Internal error %d, cannot connect to experiment\n", status);
  if (status != CM_SUCCESS)
    return 1;

  if (display_period)
    printf("OK\n");

  /* shutdown previous frontend */
  status = cm_shutdown(frontend_name);
  if (status == CM_SHUTDOWN && display_period)
    {
    printf("Previous frontend stopped successfully.\n");
    ss_sleep(1000);
    }

  if (cm_register_transition(TR_START, tr_start) != CM_SUCCESS ||
      cm_register_transition(TR_PRESTOP, tr_prestop) != CM_SUCCESS ||
      cm_register_transition(TR_PAUSE, tr_pause) != CM_SUCCESS ||
      cm_register_transition(TR_RESUME, tr_resume) != CM_SUCCESS)
    {
    printf("Failed to start local RPC server");
    cm_disconnect_experiment();
    free(event_buffer);
    return 1;
    }

  cm_get_experiment_database(&hDB, &status);

  /* set time from server */
  cm_synchronize(NULL);

  /* turn on keepalive messages with increased timeout */
  cm_set_watchdog_params(TRUE, 40000);

  /* turn off watchdog if in debug mode */
  if (debug)
    cm_set_watchdog_params(TRUE, 0);

  /* increase RPC timeout to 60sec for logger with exabyte */
  rpc_set_option(0, RPC_OTIMEOUT, 60000);

  /* set own message print function */
  if (display_period)
    cm_set_msg_print(MT_ALL, MT_ALL, message_print);

  /* call user init function */
  if (display_period)
    printf("Init hardware...");
  if (frontend_init() != CM_SUCCESS)
    {
    if (display_period)
      printf("\n");
    cm_disconnect_experiment();
    free(event_buffer);
    return 1;
    }

  /* reqister equipment in ODB */
  register_equipment();

  if (display_period)
    printf("OK\n");

  /* initialize screen display */
  if (display_period)
    {
    ss_sleep(1000);
    display(TRUE);
    }

  /* call main scheduler loop */
  status = scheduler();

  /* call user exit function */
  frontend_exit();

  /* close slow control drivers */
  for (i=0 ; equipment[i].eq_name[0] ; i++)
    if (equipment[i].eq_info.eq_type == EQ_SLOW &&
        equipment[i].status == FE_SUCCESS)
      equipment[i].func(CMD_EXIT, 0);

  /* close network connection to server */
  cm_disconnect_experiment();

  if (display_period)
    {
    ss_clear_screen();
    if (status == RPC_SHUTDOWN)
      ss_printf(0, 0, "Frontend shut down.");
    else
      ss_printf(0, 0, "Network connection aborted.");
    ss_printf(0, 1, "");
    }

  free(event_buffer);

  return 0;
}
